/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 **/

/**
 * This module wraps the nmap commands used for IP devices' interfaces binds (mac-ip-hostname)
 * Created by Iacopo Papalini <iacopo@domotz.com> on 18/05/16.
 */

const MODULE_PREFIX = 'IP_DISCOVERY - ';

module.exports.factory = function (q, proc, hostInfo, nmapCommands, nmapXml2Js, fastParser, settingsManager, log, lodash) {

    "use strict";
    var iteration = 0;

    var my_console = log.decorateLogs(MODULE_PREFIX + iteration + '  ');
    var SPAWN_HARD_TIMEOUT = 180000;

    function setIteration(i) {
        iteration = i;
        my_console = log.decorateLogs(MODULE_PREFIX + iteration + '  ');
    }

    function executeDiscoveryCommands(commandList) {
        my_console.info('executeDiscoveryCommands START ');

        var promises = [];
        commandList.forEach(function (command) {
            promises.push(proc.exec(command, {timeout: SPAWN_HARD_TIMEOUT}).then(nmapXml2Js.toJSON).then(fastParser.parse));
        });
        return q.allSettled(promises);
    }

    function getFulfilledPromisesResults(results) {
        /**
         * see
         * https://github.com/kriskowal/q/wiki/API-Reference#promiseallsettled
         * for information about the structure returned by allSettled
         * This function returns only the device bindings for the discoveries that
         * actually succeeded.
         */
        var fulfilled = lodash.map(results, function (netScanResult) {
            if (netScanResult.state === 'fulfilled') {
                return netScanResult.value;
            }
        });
        return lodash.filter(fulfilled, function (res) {
            return res !== null;
        });
    }

    function getAgentHostname() {
        var tenant = process.env.TENANT;
        if (tenant === 'adi') {
            return 'Capture';
        }
        return 'Domotz';
    }

    /**
     * Discovers all the IP devices in the given interfaces. When done the callback is invoked with, as argument,
     * an Array of Object similar to:
     * {
     *   'mac': '[the mac address]',
     *   'ip': '[the ip address]',
     *   'host_name': '[the host name]'
     * }
     */
    function discoverAll() {
        var interfaces = hostInfo.getInterfacesList();
        var name;
        var commandList = [];
        var myInterfacesBindings = {};
        my_console.info('Discover ALL START - Interfaces: %s', JSON.stringify(interfaces));

        if (settingsManager.isBroadcastDiscoveryDisabled()) {
            my_console.info('Discover ALL START, broadcast discovery is disabled, scan only known devices');
        }
        for (name in interfaces) {
            if (interfaces.hasOwnProperty(name)) {
                var iface = interfaces[name];
                var ip = iface.ip;
                var mac = iface.mac;
                var netmask = iface.netmask;
                var nmapType = nmapCommands.NMAP_TYPE.DEFAULT;
                if (mac === null) {
                    nmapType = nmapCommands.NMAP_TYPE.PING;
                }
                if (!iface.ip) {
                    my_console.warn('Discover ALL, missing ip address for interface: %s', name);
                    continue;
                }
                if (nmapType == nmapCommands.NMAP_TYPE.PING) {
                    my_console.info('Custom Interface for Monitoring External Network: %s/%s', ip, netmask);
                } else {
                    my_console.info('Real Interface for Monitoring Internal Network %s/%s. Add Binding: %s --> %s', ip, netmask, mac, ip);
                    var macUpperCase = mac.toUpperCase();
                    if (!myInterfacesBindings.hasOwnProperty(macUpperCase)) {
                        myInterfacesBindings[macUpperCase] = [];
                    }
                    myInterfacesBindings[macUpperCase].push({ip: ip, host_name: getAgentHostname()});
                }
                var commands;
                if (!settingsManager.isBroadcastDiscoveryDisabled()) {
                    commands = nmapCommands.getFastNmapCmd(ip, netmask, nmapType);
                    commandList = commandList.concat(commands);
                } else if (iface.ipList.length > 0) {
                    commandList = commandList.concat([nmapCommands.getRetryCmd(iface.ipList, nmapType)]);
                }

            }
        }

        return executeDiscoveryCommands(commandList).then(function (results) {
            my_console.info('executeDiscoveryCommands END');

            var devices = getFulfilledPromisesResults(results);
            my_console.info('Discover ALL END - Interfaces: %s', JSON.stringify(interfaces));

            devices.push(myInterfacesBindings);
            return devices;
        });

    }

    function discover(ipList, nmapType, hostTimeout) {
        my_console.info('Discover RETRY called - IP addresses: %s', JSON.stringify(ipList));
        var commandList = [nmapCommands.getRetryCmd(ipList, nmapType, hostTimeout)];
        return executeDiscoveryCommands(commandList).then(function (results) {
            my_console.info('executeDiscoveryCommands END ');
            my_console.info('Discover RETRY END - IP addresses: %s', JSON.stringify(ipList));
            return getFulfilledPromisesResults(results);
        });
    }

    function assureInterfaces() {
        var deferred = q.defer();
        try {
            var interfaces = hostInfo.getInterfacesList();
            if (Object.keys(interfaces).length === 0) {
                my_console.info('assureInterfaces - No interfaces available, trying to rediscover host info');
                hostInfo.recoverHostInfo().finally(
                    function () {
                        deferred.resolve();
                    });
            } else {
                deferred.resolve();
            }
        }
        catch(err){
            my_console.info('assureInterfaces - Error trying to get interfaces list %s', err);
            deferred.resolve();
        }
        return deferred.promise;
    }


    return {
        setIteration: setIteration,
        discoverAll: discoverAll,
        discover: discover,
        assureInterfaces: assureInterfaces
    };
};





