/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Iacopo Papalini <iacopo@domotz.com>.
 */

var MIN_SCANNING_INTERVAL_IN_SECONDS = 3* 60 * 60; // 3 hr min between scans

module.exports.ScanSample = function (mac) {
    var time = null;

    return {
        mac: mac,
        ips: null,
        time: time,
        payload: null,
        start: function () {
            this.time = new Date();
        },
        signalError: function () {
            this.time = null;
        },
        stillValid: function () {
            if (this.time === null) {
                return false;
            }
            var diffInSeconds = (new Date().getTime() - this.time.getTime()) / 1000;
            return diffInSeconds < MIN_SCANNING_INTERVAL_IN_SECONDS;
        }
    };
};

/**
 * Compacts the list of fresh interface bindings in an array with mac addresses and lists of IP addresses.
 *
 * See spec for example.
 * @param interfacesBindingStorage
 * @returns {[]}
 */
module.exports.indexBindingStorageByMac = function (interfacesBindingStorage) {
    var ret = [];
    var tmp = {};
    var devices = interfacesBindingStorage.listFresh();
    devices.forEach(function (dev) {
        if (dev.ip === null) {
            return;
        }
        if (tmp[dev.mac] === undefined) {
            tmp[dev.mac] = [];
        }
        tmp[dev.mac].push(dev.ip);
    });

    for (var mac in tmp) {
        if (tmp.hasOwnProperty(mac)) {
            ret.push({mac: mac, ips: tmp[mac]});
        }
    }
    return ret;
};

var Context = function (scanSample, myConsole, resourceLocator, callback) {
    return {
        scanSample: scanSample,
        myConsole: myConsole,
        resourceLocator: resourceLocator,
        callback: callback
    };
};

/**
 * Builds an asynchronous function that performs the scan and send the result to the engine if needed.
 * @returns {scanAndReport}
 */
module.exports.scanAndReportBuilder = function (scanSample, resourceLocator, logger) {
    function scanAndReport(callback) {
        var nmapCommands = resourceLocator.networkTools.nmapCommands;
        var myConsole = resourceLocator.log.decorateLogs(scanSample.mac + ' - ', logger);

        scanSample.start();
        var context = Context(scanSample, myConsole, resourceLocator, callback);

        var exec = resourceLocator.utils.childProcess.exec;
        var cmd = nmapCommands.getPortNmapCmdMultiIp(scanSample.ips);
        exec(cmd, {}, module.exports.parseNmapResultBuilder(context));
    }

    return scanAndReport;
};


module.exports.parseNmapResultBuilder = function (context) {
    var xml2js = context.resourceLocator.xml2js;
    var myConsole = context.myConsole;
    var callback = context.callback;
    var scanSample = context.scanSample;

    return function (error, stdout) {
        if (error) {
            myConsole.error("Error during nmap execution: %s", error);
            return callback();
        }
        try {
            xml2js.parseString(stdout, module.exports.processJsonAndSendPayloadBuilder(context));
        }
        catch (error) {
            scanSample.signalError();
            myConsole.error(error);
            return callback();
        }
    };
};


module.exports.processJsonAndSendPayloadBuilder = function (context) {
    var myConsole = context.myConsole;
    var callback = context.callback;
    var scanSample = context.scanSample;
    var resourceLocator = context.resourceLocator;
    var parser = require('./parser');

    return function (error, result) {
        if (error) {
            myConsole.error("Cannot parse nmap XML");
            return callback();
        }
        var payload;
        try {
            payload = parser(result, myConsole);
        }
        catch (error) {
            myConsole.error(error.message);
            return callback();
        }

        function onSuccess() {
            myConsole.info("Payload successfully sent to engine");
            return callback();
        }

        function onError() {
            myConsole.info("Cannot send payload to engine, resetting it");
            scanSample.setPayload(null);
            return callback();
        }

        resourceLocator.engineRequest.cachedSendWithPromise('/subnet/ip/device/' + scanSample.mac + '/open-port', 'PUT', payload).then(onSuccess).catch(onError);
    };
};