/** This file is part of Domotz Agent.
 * Copyright (C) 2016  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Iacopo Papalini <iacopo@domotz.com> on 08/09/16.
 *
 * You need fing installed on the device in order to enable the device
 * network jamming
 */
const JAMMED_FILE_NAME = "jammed.txt";
const LOG_PREFIX = "FING_JAM_DEVICES - ";
const FING_LIB_PATH = "/usr/local/lib/fing";
const FING_BIN_DEFAULT_PATH = "/usr/local/lib/fing/fing.bin";
const FING_ENV_OPTIONS = {
    env: {
        'LD_LIBRARY_PATH': FING_LIB_PATH,
        'DYLD_LIBRARY_PATH': FING_LIB_PATH
    }
};

var path = require('path');

var jamDevicesFileManager = function (macListFile, myConsole) {
    "use strict";
    var fs = require('fs');
    var previousMacAddresses = null;
    var internetPauseAvailable = null;
    var INTERNET_PAUSE_SUFFIX = ',internet-pause';

    function listChanged(macAddresses) {
        if (!previousMacAddresses) {
            return true;
        }
        if (macAddresses.length !== previousMacAddresses.length) {
            return true;
        }
        else {
            for (var i = macAddresses.length; i-- > 0;) {
                if (macAddresses[i] !== previousMacAddresses[i]) {
                    return true;
                }
            }
        }
        return false;
    }

    function writeFile(contents) {
        myConsole.debug('Synchronously writing file %s', macListFile);
        try {
            fs.writeFileSync(macListFile, contents + '\n');
            myConsole.debug('File written %s (%s)', macListFile, JSON.stringify(contents));
            return contents.length > 0;
        } catch (e) {
            myConsole.error("Cannot write file %s, killing fing", macListFile);
            return false;
        }
    }

    function writeFileIfListChanged(macAddresses) {
        /**
         * Using the macAddresses list, writes its contents in the macListFile.
         * @returns boolean true if the fing process should run after the update
         */
        if (listChanged(macAddresses)) {
            myConsole.info("Jammed device list changed");
        } else {
            return macAddresses.length > 0;
        }
        previousMacAddresses = macAddresses;

        var contents = '';
        if (internetPauseAvailable && macAddresses.length > 0) {
            contents = macAddresses.join(INTERNET_PAUSE_SUFFIX + '\n') + INTERNET_PAUSE_SUFFIX;
        } else {
            contents = macAddresses.join('\n');
        }
        return writeFile(contents);
    }

    function setInternetPauseAvailable(value) {
        myConsole.info("Setting internet pause availability as: %s", value);
        internetPauseAvailable = value;
    }

    function isInternetPauseAvailable() {
        /*
         * returns null if the information is not available yet,
         * true for versions >= 4.x, false otherwise
         *
         * */
        return internetPauseAvailable;
    }

    return {
        write: writeFileIfListChanged,
        setInternetPauseAvailable: setInternetPauseAvailable,
        isInternetPauseAvailable: isInternetPauseAvailable
    };
};


var fingProcessManager = function (fingCommand, macListFile, childProcess, procManager, q, myConsole) {
    "use strict";
    var fingProcess = null;
    var fingParameters = ["--kickout", '-', '--kocfg', macListFile];

    function isFingRunning() {
        if (fingProcess !== null) {
            myConsole.debug("Fing process is running");
            return true;
        } else {
            myConsole.debug("Fing process not existing");
            return false;
        }
    }

    function assureFingIsRunning() {
        if (!isFingRunning()) {
            myConsole.info('Starting fing');

            fingProcess = childProcess.spawn(fingCommand, fingParameters, {env: getEnrichedEnv()});

            fingProcess.on('close', function () {
                myConsole.info("Fing process killed");
                fingProcess = null;
            });

            fingProcess.stderr.on('data', function (data) {
                myConsole.warn("ERR DATA  %s", data);
            });

            fingProcess.stdout.on('data', function (data) {
                myConsole.verbose("STDOUT DATA " + data);
            });

            fingProcess.on('error', function (err) {
                myConsole.warn('Fing block call: unable to execute Fing: %s', JSON.stringify(err));
            });

        } else {
            myConsole.debug("Fing already running, not executed");
        }
    }

    function assureFingIsNotRunning() {
        if (isFingRunning()) {
            myConsole.info('Killing fing');
            fingProcess.kill('SIGKILL');
            fingProcess = null;
        } else {
            myConsole.info("Fing not running, not killing it");
        }
    }

    function getEnrichedEnv() {
        var enrichedEnv = JSON.parse(JSON.stringify(process.env));
        enrichedEnv.LD_LIBRARY_PATH = (enrichedEnv.LD_LIBRARY_PATH ? enrichedEnv.LD_LIBRARY_PATH + ":" : "") + process.env.DOMOTZ_LIB_DIR + ":" + FING_ENV_OPTIONS.env.LD_LIBRARY_PATH;
        enrichedEnv.DYLD_LIBRARY_PATH = (enrichedEnv.DYLD_LIBRARY_PATH ? enrichedEnv.DYLD_LIBRARY_PATH + ":" : "") + process.env.DOMOTZ_LIB_DIR + ":" + FING_ENV_OPTIONS.env.LD_LIBRARY_PATH;
        return enrichedEnv;
    }

    function getFingVersion() {
        var deferred = q.defer();
        getEnrichedEnv();
        procManager.exec(fingCommand + " --version", {env: getEnrichedEnv()}).then(function (stdout) {
            var trimmed = stdout.trim();
            if (!trimmed.match(/\d+\.\d+\.\d+/)) {
                myConsole.warn("fing version is malformed: %s", trimmed);
                deferred.reject();
                return;
            }
            deferred.resolve(trimmed);
        }).catch(function () {
            deferred.reject();
        });
        return deferred.promise;
    }

    return {
        assureFingIsRunning: assureFingIsRunning,
        assureFingIsNotRunning: assureFingIsNotRunning,
        getFingVersion: getFingVersion
    };
};


var fingWrapper = function (childProcess, engineRequest, procManager, q, log) {
    "use strict";
    var myConsole = log.decorateLogs(LOG_PREFIX);
    var env = process.env;
    var macListFile = path.join((env.DOMOTZ_CONF_DIR), JAMMED_FILE_NAME);
    var fingCommand = (env.FING_BINARY_FULL_PATH || FING_BIN_DEFAULT_PATH);
    var fileWriter = jamDevicesFileManager(macListFile, myConsole);
    var processManager = fingProcessManager(fingCommand, macListFile, childProcess, procManager, q, myConsole);

    function handleServerResponse(data) {
        var shouldRun = fileWriter.write(data);
        if (shouldRun && data.length > 0) {
            processManager.assureFingIsRunning();
        }
        else {
            processManager.assureFingIsNotRunning();
        }
    }

    function _updateMacAddresses() {
        myConsole.info("Updating mac addresses list");

        engineRequest.sendWithPromise('/device/jammed', 'GET').then(function (data) {
            handleServerResponse(data);
        }).catch(function (error) {
            myConsole.error('Cannot update list, assuming no device to jam, %s', JSON.stringify(error));
            if (error.status.toString()[0] == '4') {
                myConsole.error("Client error %s, set jammed list as empty", error.status);
                handleServerResponse([]);
            } else {
                myConsole.error("Server error %s, leave jammed list untouched", error.status);
            }
        });
    }

    return {
        updateMacAddresses: function () {
            if (fileWriter.isInternetPauseAvailable() === null) {

                processManager.getFingVersion().then(function (version) {
                    myConsole.info("Fing version is %s", version);
                    if (parseInt(version.split('.')[0]) >= 4) {
                        fileWriter.setInternetPauseAvailable(true);
                    } else {
                        myConsole.info("Fing version %s does not support Internet Pause", version);
                        fileWriter.setInternetPauseAvailable(false);
                    }
                    _updateMacAddresses();
                }).catch(function () {
                    myConsole.warn("Unable to get Fing version, Internet Pause feature will NOT be available");
                    _updateMacAddresses();
                });
            } else {
                _updateMacAddresses();
            }
        },
        killDeviceJammer: function () {
            processManager.assureFingIsNotRunning();
        },
        getFingVersion: processManager.getFingVersion
    };
};

module.exports.fingWrapper = fingWrapper;

// For testing purposes only
module.exports._fileManager = jamDevicesFileManager;
module.exports._fingProcessManager = fingProcessManager;
