/** This file is part of Domotz Agent.
 * Copyright (C) 2017  Domotz Ltd
 *
 * Domotz Agent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Domotz Agent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Domotz Agent.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Created by Massimiliano Cuzzoli <massi@domotz.com> on 30/10/17.
 *
 */

var jayson = require('jayson');
var fs = require('fs');

const LOG_PREFIX = "LUXUL_JAM_DEVICES - ";
const DEFAULT_ROUTER_IP = '192.168.0.1'
const DEFAULT_PASSWORD = "domotz"
const DOMOTZ_API_PORT = 8008
const DOMOTZ_API_PATH = 'ubus'
const GET_SESSION_COMMAND = 'login'
const BLOCK_COMMAND = 'block'
const UNBLOCK_COMMAND = 'unblock'
const GET_BLOCKED_MAC_COMMAND = "get_blocked_macs"
const LOGIN_TOKEN = "00000000000000000000000000000000"
const API_VERSION = 2
const UBUS_USERNAME = "domotz"
const UBUS_PASSWD_FILE = "/etc/domotz-password.txt"
const UBUS_IP_FILE = "/etc/luxul_api_ip"

var jamStateManager = function () {
    var disappearedMacAddresses = [];
    var appearedMacAddresses = [];

    var updateBlockingFeatureState = function (blockedMacAddresses, macAddressesToBlock) {
        disappearedMacAddresses = blockedMacAddresses.filter(function (i) {
            return macAddressesToBlock.indexOf(i) < 0;
        });
        appearedMacAddresses = macAddressesToBlock.filter(function (i) {
            return blockedMacAddresses.indexOf(i) < 0;
        });

        return (disappearedMacAddresses.length > 0 || appearedMacAddresses.length > 0);
    };

    return {
        update: updateBlockingFeatureState,
        getMacAddressesToUnblock: function () {
            return disappearedMacAddresses;
        },
        getMacAddressesToBlock: function () {
            return appearedMacAddresses;
        }
    }
};

var luxulUbusManager = function (q) {
    var router_ip = router_ip ? router_ip : "http://"+(read_luxul_file_sync(UBUS_IP_FILE) || DEFAULT_ROUTER_IP);
    var pass = pass ? pass : (read_luxul_file_sync(UBUS_PASSWD_FILE) || DEFAULT_PASSWORD);

    "use strict";
    function read_luxul_file_sync(file) {

        var content = null;
        try {
            var row = fs.readFileSync(file, "utf8");
            content = row.toString().split('\n').map(function (line) {
                return line.trim();
            }).filter(Boolean);
        } catch (error) {
            console.error("%s:Error reading file %s stack_trace: %s", LOG_PREFIX, file, JSON.stringify(error));
        }
        return content;
    }

    function message_builder(token, topic, command, command_parameters) {
        var params_content = [token, topic, command, command_parameters]
        var message = {
            "id": 1,
            "method": "call",
            "params": params_content
        }
        return message;
    }

    function requestToLuxulWithDeferred(message, deferred) {
        var url = [
            [router_ip, "8008"].join(":"), DOMOTZ_API_PATH
        ].join("/")
        var client = jayson.client.http(url);
        client.request(message.method, message.params, message.id, function (error, response) {
            try {
                if (error) {
                    deferred.reject("error connecting to luxul ubus: " + JSON.stringify(error));
                }
                var resp = message.params[0] === LOGIN_TOKEN ? response.result[1].ubus_rpc_session : response.result
                deferred.resolve(resp);
            } catch (err) {
                deferred.reject("Exception connectiong to luxul ubus");
            }
        });
    }
    var get_session_token = function () {
        var auth = {
            "username": UBUS_USERNAME,
            "password": pass.toString()
        };
        var message = message_builder(LOGIN_TOKEN, "session", GET_SESSION_COMMAND, auth)
        var deferred = q.defer();
        requestToLuxulWithDeferred(message, deferred);
        return deferred.promise;
    };

    var get_blocked_devices = function (token) {
        var message = message_builder(token, "luxul.firewall", GET_BLOCKED_MAC_COMMAND, {})
        var deferred = q.defer();
        requestToLuxulWithDeferred(message, deferred);
        return deferred.promise;
    };

    var block_unblock_mac_address = function (token, action, mac_list) {
        var macs = {}
        if (!mac_list || !mac_list.constructor === Array || (mac_list.constructor === Array && mac_list.length == 0)) {
            console.debug("%s Empty mac list. Nothing to do", LOG_PREFIX);
            return q(null);
        }
        macs[action] = mac_list.map(function (address) {
            return address.hasOwnProperty("mac") ? address : {
                "mac": address
            };
        })
        var deferred = q.defer();
        var message = message_builder(token, "luxul.firewall", action, macs)
        requestToLuxulWithDeferred(message, deferred);
        return deferred.promise;
    };

    var handle_blocking_feature_response = function (data, actionType) {
        if (data === null || data.constructor === Array || (data.constructor === Array && data.length == 0)) {
            console.info("%s - %s : empty data response", LOG_PREFIX, actionType);
            return;
        }
        if (data.length > 0 && data.result[0] > 0) {
            throw new Error(actionType + " feature fail with error: " + JSON.stringify(data.error));
        } else {
            console.info("%s - %s feature executed - result=%s", LOG_PREFIX, actionType, JSON.stringify(data));
        }
    };

    return {
        getSessionToken: get_session_token,
        getBlockedDevices: get_blocked_devices,

        blockDeviceList: function (mac_list) {
            return get_session_token().
            then(function (token) {
                block_unblock_mac_address(token, BLOCK_COMMAND, mac_list).
                then(function (data) {
                    handle_blocking_feature_response(data, "Block")
                })
            });
        },

        unblockDeviceList: function (mac_list) {
            return get_session_token().
            then(function (token) {
                block_unblock_mac_address(token, UNBLOCK_COMMAND, mac_list).
                then(function (data) {
                    handle_blocking_feature_response(data, "Unblock")
                })
            });
        },

        unblockAllDevices: function (mac_list) {
            return get_session_token().
            then(function (token) {
                get_blocked_devices(token).
                then(function (ubusResponse) {
                    if (ubusResponse && ubusResponse.response === 0) {
                        return (ubusResponse.data)[0].macs;
                    }
                    throw new Error("Unable to retrieve blocked mac list " + JSON.stringify(ubusResponse));
                }).
                then(function (blockedMacAddresses) {
                    return block_unblock_mac_address(UNBLOCK_COMMAND, blockedMacAddresses);
                }).
                then(function (data) {
                    handle_blocking_feature_response(data, "UnblockAll")
                });
            });
        }
    };
};

var jamDriver = function (engineRequest, path, q) {
    "use strict";
    var blockingFeatureAvailable = null;
    var stateManager = jamStateManager();
    var ubusManager = luxulUbusManager(q);

    function handleServerResponse(data) {
        ubusManager.getSessionToken().
        then(function (token) {
            ubusManager.getBlockedDevices(token).
            then(function (ubusResponse) {
                if (ubusResponse && ubusResponse[1] && ubusResponse[1].blocked) {
                    var blockedDevices = ubusResponse[1].blocked;
                    console.debug("%s Already Blocked MAC #%d", LOG_PREFIX, blockedDevices.length);
                    console.debug("%s MAC list to block : #%d", LOG_PREFIX, data.length);
                    return stateManager.update(blockedDevices, data);
                } else {
                    throw new Error("Unable to retrieve blocked mac list " + JSON.stringify(ubusResponse));
                }
            }).
            then(function (isActionRequired) {
                console.debug("%s is action required : %s", LOG_PREFIX, isActionRequired);
                if (isActionRequired) {
                    return ubusManager.unblockDeviceList(stateManager.getMacAddressesToUnblock()).
                    then(ubusManager.blockDeviceList(stateManager.getMacAddressesToBlock()));
                }
            }).
            catch(function (error) {
                console.error("%s Error on device jamming : %s", LOG_PREFIX, JSON.stringify(error));
            });
        })
    }

    function updateMacAddresses() {
        console.info(LOG_PREFIX + "Updating mac addresses list");
        engineRequest.sendWithPromise('/device/jammed', 'GET').
        then(function (data) {
            handleServerResponse(data);
        }).
        catch(function (error) {
            console.error('%s Cannot update list, assuming no device to jam, %s', LOG_PREFIX, JSON.stringify(error));
            if (error.status.toString()[0] === '4') {
                console.error("%s Client error %s, set jammed list as empty", LOG_PREFIX, error.status);
                handleServerResponse([]);
            } else {
                console.error("%s Server error %s, leave jammed list untouched", LOG_PREFIX, error.status);
            }
        });
    }

    return {
        updateMacAddresses: function () {
            if (blockingFeatureAvailable === true) {
                updateMacAddresses();
                return;
            }

            if (blockingFeatureAvailable === null) {
                try {
                    blockingFeatureAvailable = true;
                    updateMacAddresses();
                } catch (error) {
                    blockingFeatureAvailable = false
                }
            }
        },
        killDeviceJammer: function () {
            return ubusManager.unblockAllDevices().catch(function (error) {
                console.error("%s Error on unblock all devices : %s", LOG_PREFIX, error)
            });
        },
        isBlockingFeatureAvailable: function () {
            return blockingFeatureAvailable;
        },
        luxulApiVersion: function () {
            return API_VERSION;
        }
    };
};

module.exports.jamDriver = jamDriver;

// For testing purposes only
module.exports._stateManager = jamStateManager;
module.exports._ubusManager = luxulUbusManager;